
import sys
import inspect
import requests
from urllib.parse import urlparse
from unittest.mock import MagicMock, patch, create_autospec
from odoo import api
from odoo.tests import common, tagged
from odoo.addons.saas_worker import metabase
from odoo.addons.saas_worker.util import version_gte

try:
    from odoo.addons.saas_trial.tests.common import TestTrialCommon
except ImportError:
    class TestTrialCommon(common.BaseCase):
        @classmethod
        def setUpClass(cls):
            cls.skipTest(cls, "Skipping test, TestTrialCommon not available")


@tagged('post_install', '-at_install', 'post_saas_trial_creation')
class TestPlanChangeWarnings(TestTrialCommon):
    """Test that we warn user when it's going to change/upgrade their subscription plan."""

    def setUp(self):
        super().setUp()
        self.db = metabase.get(self.env.cr.dbname)

    def _mock_method_autospec(self, method, *args, **kwargs):
        """Mock a method with autospec=True while keeping same api decorator and method signature."""
        api_decorator = {
            'model': api.model,
            'model_create': api.model_create_multi,
        }.get(getattr(method, '_api', None))

        method_mock = create_autospec(method, *args, **kwargs)
        if api_decorator:
            method_mock = api_decorator(method_mock)
        if sys.version_info < (3, 8):
            # backport bpo-17185: set __signature__ so that getargspec() used by
            # `odoo.api.split_context()` for version <= saas~12.1 works properly;
            # without the real signature, calling this method through `call_kw()`
            # will crash.
            method_mock.__signature__ = inspect.signature(method)
        return method_mock

    def _patch_model_method(self, model, method_name, *args, **kwargs):
        model_klass = type(self.env[model])
        method = getattr(model_klass, method_name)
        mock_method = self._mock_method_autospec(method, *args, **kwargs)
        return patch.object(model_klass, method_name, new=mock_method)

    def _set_metabase_mode_for_plan(self, plan):
        metabase_mode = 'paid' if plan in ('standard', 'custom') else 'trial'
        metabase_extra_apps = plan != '1appfree'
        metabase.update(self.db, mode=metabase_mode, extra_apps=metabase_extra_apps)

    def _mock_post_accounts_get_pricing(self, pricing=False):
        assert pricing in (False, '1appfree', 'trial', 'standard', 'custom')
        _request_post = requests.post
        _request_mock_getpricing = MagicMock()
        _request_mock_getpricing.return_value.json.return_value = {'result': pricing}

        def _request_post_mock(url, *args, **kwargs):
            urlp = urlparse(url)
            if urlp.path == "/odoo-enterprise/get-pricing":
                return _request_mock_getpricing()  # (url, *args, **kwargs)
            return _request_post(url, *args, **kwargs)

        return patch('requests.post', wraps=_request_post_mock)

    def test_user_install_studio_warning(self):
        """Test that we ask for confirmation before installing studio, for all plans except `custom`"""
        if not version_gte("15.0"):
            self.skipTest("Studio promote dialog is only available on Odoo 15.0+")
        if self.env.ref("base.module_web_studio").state in ("to install", "installed", "to upgrade"):
            self.skipTest("Studio is already installed, skipping test")

        def _patch_module_install():
            return self._patch_model_method(
                "ir.module.module", "button_immediate_install",
                return_value=True
            )

        plans_tour = {
            # Plan:     (Tour, expected_account_get_pricing_calls)
            '1appfree': ('trial_user_install_studio_warning', 0),
            'trial': ('trial_user_install_studio_warning', 0),
            'standard': ('trial_user_install_studio_warning', 1),
            'custom': ('trial_user_install_studio_no_warning', 1),
        }
        for plan, (test_tour, expected_account_get_pricing_calls) in plans_tour.items():
            with self.subTest(plan=plan):
                self._set_metabase_mode_for_plan(plan)

                with self._mock_post_accounts_get_pricing(pricing=plan) as mock_accounts_get_pricing, \
                        _patch_module_install() as mock_module_install:
                    self.start_tour('/web', test_tour, login=self.user_admin.login)
                    self.assertEqual(
                        mock_accounts_get_pricing.call_count, expected_account_get_pricing_calls,
                    )
                    self.assertEqual(mock_module_install.call_count, 1)

    def test_user_invite_from_settings_dashboard(self):
        """Test that we ask for confirmation before inviting users, for all plans except `1appfree`"""

        if version_gte("saas~12.5"):
            web_create_users_method_name = "web_create_users"
        else:
            web_create_users_method_name = "web_dashboard_create_users"

        def _patch_web_create_users():
            return self._patch_model_method("res.users", web_create_users_method_name, return_value=True)

        plans_tour = {
            '1appfree': 'trial_user_invite_from_setting_dashboard_no_warning',
            'trial': 'trial_user_invite_from_setting_dashboard_warning',
            'standard': 'trial_user_invite_from_setting_dashboard_warning',
            'custom': 'trial_user_invite_from_setting_dashboard_warning',
        }
        for plan, test_tour in plans_tour.items():
            with self.subTest(plan=plan):
                self._set_metabase_mode_for_plan(plan)

                with self._mock_post_accounts_get_pricing(pricing=plan), \
                        _patch_web_create_users() as mock_web_create_users:
                    self.start_tour('/web', test_tour, login=self.user_admin.login)
                    self.assertEqual(mock_web_create_users.call_count, 1)

    def test_settings_install_module_warning(self):
        """Test that we ask for confirmation before enabling a "module" option in settings, for free & trial plans"""
        warning_trigger = "on_save" if version_gte("16.0") else "on_change"
        test_tour_warning = f"trial_settings_install_module_{warning_trigger}_warning"
        test_tour_no_warning = f"trial_settings_install_module_{warning_trigger}_no_warning"

        # For the purpose of the test, set "auth_oauth" as being an application
        # to allow testing as this module is always present on the settings form
        self.env.ref("base.module_auth_oauth").write({
            'application': True,
            'state': 'uninstalled',
        })

        def _patch_res_config_settings_execute():
            return self._patch_model_method("res.config.settings", "execute", return_value=True)

        plans_tour = {
            '1appfree': test_tour_warning,
            'trial': test_tour_warning,
            'standard': test_tour_no_warning,
            'custom': test_tour_no_warning,
        }
        for plan, test_tour in plans_tour.items():
            with self.subTest(plan=plan):
                self._set_metabase_mode_for_plan(plan)

                with self._mock_post_accounts_get_pricing(pricing=plan), \
                        _patch_res_config_settings_execute() as mock_settings_execute:
                    self.start_tour("/web", test_tour, login=self.user_admin.login)
                    self.assertEqual(mock_settings_execute.call_count, 1)

    def test_apps_install_module_warning(self):
        """Test that we ask for confirmation before installing a module in apps (kanban/form), for free & trial plans"""
        test_tour_kanban = "trial_apps_install_module_kanban"
        test_tour_form = "trial_apps_install_module_form"
        test_tours_warning = {
            'kanban': test_tour_kanban + "_warning",
            "form": test_tour_form + "_warning"
        }
        test_tours_no_warning = {
            'kanban': test_tour_kanban + "_no_warning",
            "form": test_tour_form + "_no_warning"
        }

        def _patch_module_install():
            return self._patch_model_method(
                "ir.module.module", "button_immediate_install",
                # we want the webclient to stay on the same page for the tour,
                # and webclient expect a valid action (can't return True), so
                # we're using a "close" action as a "no-op" here
                return_value={'type': 'ir.actions.act_window_close'},
            )

        plans_tour = {
            '1appfree': test_tours_warning,
            'trial': test_tours_warning,
            'standard': test_tours_no_warning,
            'custom': test_tours_no_warning,
        }
        for plan, test_tours in plans_tour.items():
            for view, test_tour in test_tours.items():
                with self.subTest(plan=plan, view=view):
                    self._set_metabase_mode_for_plan(plan)

                    with self._mock_post_accounts_get_pricing(pricing=plan), \
                            _patch_module_install() as mock_module_install:
                        self.start_tour("/web", test_tour, login=self.user_admin.login)
                        self.assertEqual(mock_module_install.call_count, 1)

    def test_company_creation_warning(self):
        """Test that we ask for confirmation before creating a new company, for trial and standard plans"""
        plans_tour = {
            '1appfree': 'trial_company_creation_no_warning',
            'trial': 'trial_company_creation_warning',
            'standard': 'trial_company_creation_warning',
            'custom': 'trial_company_creation_no_warning',
        }

        def patch_res_company_create():
            return self._patch_model_method(
                "res.company", "create", return_value=self.env.user.company_id
            )

        for plan, test_tour in plans_tour.items():
            with self.subTest(plan=plan):
                self._set_metabase_mode_for_plan(plan)

                with self._mock_post_accounts_get_pricing(pricing=plan), \
                        patch_res_company_create() as mock_res_company_create:
                    self.start_tour('/web', test_tour, login=self.user_admin.login)
                    self.assertEqual(mock_res_company_create.mock.call_count, 1)

    def test_heldesk_team_option_warning(self):
        """Test that we ask for confirmation before enabling a "module" option on a helpdesk.team, for free & trial plans"""

        if not self.env.ref("base.module_helpdesk").state in ("to install", "installed", "to upgrade"):
            self.skipTest("Helpdesk module is not installed, skipping")

        warning_trigger = "on_save" if version_gte("16.0") else "on_change"
        test_tour_warning = f"trial_helpdesk_team_install_module_{warning_trigger}_warning"
        test_tour_no_warning = f"trial_helpdesk_team_install_module_{warning_trigger}_no_warning"

        # For the purpose of the test, set helpdesk's "eLearning" and dep. as uninstalled
        self.env.ref('base.module_website_helpdesk_slides').state = 'uninstalled'
        self.env.ref('base.module_website_slides').state = 'uninstalled'

        def _patch_module_install():
            return self._patch_model_method(
                "ir.module.module", "button_immediate_install",
                return_value=True
            )

        plans_tour = {
            '1appfree': test_tour_warning,
            'trial': test_tour_warning,
            'standard': test_tour_no_warning,
            'custom': test_tour_no_warning,
        }
        for plan, test_tour in plans_tour.items():
            with self.subTest(plan=plan):
                self._set_metabase_mode_for_plan(plan)

                with self._mock_post_accounts_get_pricing(pricing=plan), \
                        _patch_module_install() as mock_module_install:
                    self.start_tour("/web", test_tour, login=self.user_admin.login)
                    self.assertTrue(mock_module_install.call_count)

    def test_apps_install_base_automation_warning(self):
        """ Test that we ask for confirmation before installing
        a base_automation in apps (kanban/form), for standard plans in 17.0+
        """

        if not version_gte("17.0"):
            self.skipTest("The pricing plan change for base.automation only applies since Odoo 17.0+")

        base_automation = self.env['ir.module.module'].search([('name', '=', 'base_automation')])
        if base_automation and base_automation.state == "installed":
            self.skipTest("Automated Rules module is already installed, skipping")

        def _patch_module_install():
            return self._patch_model_method(
                "ir.module.module", "button_immediate_install",
                # we want the webclient to stay on the same page for the tour,
                # and webclient expect a valid action (can't return True), so
                # we're using a "close" action as a "no-op" here
                return_value={'type': 'ir.actions.act_window_close'},
            )

        plans_tour = {
            '1appfree': 'trial_apps_install_base_automation_no_warning',
            'trial': 'trial_apps_install_base_automation_no_warning',
            'standard': 'trial_apps_install_base_automation_warning',
            'custom': 'trial_apps_install_base_automation_no_warning',
        }
        for plan, test_tour in plans_tour.items():
            with self.subTest(plan=plan):
                self._set_metabase_mode_for_plan(plan)

                with self._mock_post_accounts_get_pricing(pricing=plan), \
                        _patch_module_install() as mock_module_install:
                    self.start_tour("/web", test_tour, login=self.user_admin.login)
                    self.assertEqual(mock_module_install.call_count, 1)
